import argparse
import json
import sys
import os
from subprocess import check_call, CalledProcessError
import pkg_resources
from .parse.parser import WCAParser
from .sema.ast import WCARegulations, WCAGuidelines, Ruleset
from .codegen.html import WCARegulationsHtml, WCAGuidelinesHtml
from .codegen.latex import WCADocumentLatex

REGULATIONS_FILENAME = "wca-regulations.md"
GUIDELINES_FILENAME = "wca-guidelines.md"

def parse_regulations_guidelines(reg, guide):
    reg_as_str = None
    guide_as_str = None
    astreg = None
    astguide = None
    errors = []
    warnings = []
    if reg:
        with open(reg) as reg_file:
            reg_as_str = reg_file.read()
        # FIXME: do we want to just remove tabs from regs?
        reg_as_str = reg_as_str.replace("\t", "    ")
    if guide:
        with open(guide) as guide_file:
            guide_as_str = guide_file.read()
        # FIXME: do we want to just remove tabs from regs?
        guide_as_str = guide_as_str.replace("\t", "    ")
    parser = WCAParser()
    if reg_as_str:
        astreg, errors_reg, warnings_reg = parser.parse(reg_as_str, WCARegulations)
        errors.extend(errors_reg)
        warnings.extend(warnings_reg)
    if guide_as_str:
        astguide, errors_guide, warnings_guide = parser.parse(guide_as_str, WCAGuidelines)
        errors.extend(errors_guide)
        warnings.extend(warnings_guide)
    # parser.lexer.input(reg_as_str)
    # for t in parser.lexer:
        # print t
    return (astreg, astguide, errors, warnings)

def generate_html(input_regulations, input_guidelines, options):
    output_directory = options.output
    astreg, astguide, errors, warnings = parse_regulations_guidelines(input_regulations,
                                                                      input_guidelines)
    if len(errors) + len(warnings) == 0 and astreg and astguide:
        print "Compiled Regulations and Guidelines, generating html..."
        cg_reg_html = WCARegulationsHtml(options.version, options.language)
        cg_guide_html = WCAGuidelinesHtml(astreg, options.version, options.language)
        reg_html = cg_reg_html.emit(astreg)
        guide_html = cg_guide_html.emit(astguide)
        if reg_html and guide_html:
            output_reg = output_directory + "/index.html"
            output_guide = output_directory + "/guidelines.html"
            with open(output_reg, 'w+') as output_file:
                output_file.write(reg_html)
                print "Successfully written the Regulations' html to " + output_reg
            with open(output_guide, 'w+') as output_file:
                output_file.write(guide_html)
                print "Successfully written the Guidelines' html to " + output_guide
        else:
            print "Error: couldn't emit html for Regulations and Guidelines."
            sys.exit(1)
    return (errors, warnings)

def generate_latex(input_regulations, input_guidelines, options):
    output_directory = options.output
    astreg, astguide, errors, warnings = parse_regulations_guidelines(input_regulations,
                                                                      input_guidelines)
    if len(errors) + len(warnings) == 0 and astreg and astguide:
        print "Compiled Regulations and Guidelines, generating Latex..."

        languages_info = languages(False)

        cglatex = WCADocumentLatex(options.language,
                                   languages_info[options.language]["tex_encoding"])
        latex = cglatex.emit(astreg, astguide)
        if latex:
            base_filename = languages_info[options.language]["pdf"]
            output = output_directory + "/" + base_filename + ".tex"
            with open(output, 'w+') as output_file:
                output_file.write(latex)
                print "Successfully written the Latex to " + output
            if options.target == "pdf":
                latex_cmd = [languages_info[options.language]["tex_command"]]
                latex_cmd.append("-output-directory=" + output_directory)
                latex_cmd.append(output_directory + "/" + base_filename + ".tex")
                try:
                    check_call(latex_cmd)
                    # Do it twice for ToC!
                    check_call(latex_cmd)
                    print "Successfully generated pdf file!"
                    print "Cleaning temporary file..."
                    for ext in [".tex", ".aux", ".log"]:
                        to_remove = output_directory + "/" + base_filename + ext
                        print "Removing: " + to_remove
                        os.remove(to_remove)
                except CalledProcessError as err:
                    print "Error while generating pdf:"
                    print err
                    # Removing .aux file to avoid build problem
                    print "Removing .aux file"
                    os.remove(output_directory + "/" + base_filename + ".aux")
                    sys.exit(1)
                except OSError as err:
                    print "Error when running command \"" + " ".join(latex_cmd) + "\""
                    print err
                    sys.exit(1)

        else:
            print "Error: couldn't emit Latex for Regulations and Guidelines."
            sys.exit(1)
    return (errors, warnings)

def output_diff(submitted, reference):
    rules_visitor = Ruleset()
    set_submitted = rules_visitor.get(submitted)
    set_reference = rules_visitor.get(reference)
    unexpected = set_submitted - set_reference
    missing = set_reference - set_submitted
    if len(unexpected) > 0:
        print ("/!\\ These numbers are in the translation file, "
               "but not in the reference one: {%s}" % ', '.join(sorted(unexpected)))
    if len(missing) > 0:
        print ("/!\\ These numbers are in the reference file, "
               "but not in the translation one: {%s}" % ', '.join(sorted(missing)))
    return len(unexpected) + len(missing)

def generate_diff(input_ast_reg, input_ast_guide, options):
    ref_regulations, ref_guidelines = files_from_dir(options.diff)
    parse_ref = parse_regulations_guidelines(ref_regulations, ref_guidelines)
    astreg_ref, astguide_ref, errors, warnings = parse_ref
    if len(errors) + len(warnings) != 0:
        print "Couldn't compile reference Regulations and Guidelines"
    else:
        print "All checks passed!"
        diffs = 0
        if input_ast_reg and astreg_ref:
            diffs += output_diff(input_ast_reg, astreg_ref)
        elif input_ast_reg:
            print "No reference to compare the intput Regulations to"

        if input_ast_guide and astguide_ref:
            diffs += output_diff(input_ast_guide, astguide_ref)
        elif input_ast_guide:
            print "No reference to compare the input Guidelines to"
        if diffs == 0:
            print "Input file(s) and reference file(s) matched!"
        else:
            errors.append("Translation and reference did not match!")
    return (errors, warnings)


def files_from_dir(file_or_directory):
    regulations = None
    guidelines = None
    if os.path.isdir(file_or_directory):
        regulations = file_or_directory + "/" + REGULATIONS_FILENAME
        guidelines = file_or_directory + "/" + GUIDELINES_FILENAME
        if (not os.path.isfile(regulations) or
                not os.path.isfile(guidelines)):
            print ("Error: the directory '%s' must contain both the Regulations "
                   "and Guidelines files." % file_or_directory)
            sys.exit(1)
    elif os.path.isfile(file_or_directory):
        if file_or_directory.endswith(REGULATIONS_FILENAME):
            regulations = file_or_directory
        elif file_or_directory.endswith(GUIDELINES_FILENAME):
            guidelines = file_or_directory
        else:
            print "Error: couldn't detect if the input file are Regulations or Guidelines."
            sys.exit(1)
    else:
        print "Error: %s is not a file or a directory." % file_or_directory
        sys.exit(1)
    return (regulations, guidelines)

def check_output(directory):
    if not os.path.isdir(directory):
        print "Error: output is not a directory."
        sys.exit(1)

def languages(display=True):
    # Get information about languages from the config file (tex encoding, pdf filename, etc)
    languages_info = json.loads(pkg_resources.resource_string(__name__, "data/languages.json"))

    print " ".join(languages_info)
    return languages_info

def run():
    argparser = argparse.ArgumentParser()
    action_group = argparser.add_mutually_exclusive_group()
    action_group.add_argument('--target', help='Select target output kind',
                              choices=['latex', 'pdf', 'html', 'check'])
    action_group.add_argument('--diff', help='Diff against the specified file')
    argparser.add_argument('-o', '--output', default='build/', help='Output directory')
    argparser.add_argument('-l', '--language', default='english', help='Language of the file')
    argparser.add_argument('-v', '--version', default='unknown',
                           help='Git hash corresponding to the files')
    argparser.add_argument('input', help='Input file or directory')


    options = argparser.parse_args()

    input_regulations, input_guidelines = files_from_dir(options.input)

    errors = []
    warnings = []


    if not options.diff and not options.target:
        print "Nothing to do, exiting..."
        sys.exit(0)

    if options.target == "latex" or options.target == "pdf":
        check_output(options.output)
        if not input_regulations or not input_guidelines:
            print ("Error: both the Regulations and Guidelines are needed "
                   "to generate the Latex file.")
            sys.exit(1)
        errors, warnings = generate_latex(input_regulations, input_guidelines, options)
    elif options.target == "html":
        check_output(options.output)
        if not input_regulations or not input_guidelines:
            print ("Error: both the Regulations and Guidelines are needed "
                   "to generate the html file.")
            sys.exit(1)
        errors, warnings = generate_html(input_regulations, input_guidelines, options)
    elif options.target == "check" or options.diff:
        print "Checking input file(s)..."
        astreg, astguide, errors, warnings = parse_regulations_guidelines(input_regulations,
                                                                          input_guidelines)
        if len(errors) + len(warnings) == 0:
            print "All checks passed!"
            if options.diff:
                print "Checking reference file(s) for diff"
                errors, warnings = generate_diff(astreg, astguide, options)


    # If some errors or warnings have been detected, output them
    if len(errors) + len(warnings) != 0:
        print "Couldn't compile file, the following occured:"
        for err in errors:
            print " - Error: " + err
        for warn in warnings:
            print " - Warning: " + warn
        sys.exit(1)

if __name__ == '__main__':
    run()



class WCADocument(object):
    def __init__(self, title, version, sections):
        self.title = title
        self.sections = sections
        self.version = version

class WCARegulations(WCADocument):
    def __init__(self, title, version, sections):
        super(WCARegulations, self).__init__(title, version, sections)

class WCAGuidelines(WCADocument):
    def __init__(self, title, version, sections):
        super(WCAGuidelines, self).__init__(title, version, sections)

class Section(object):
    def __init__(self, title, intro, content):
        self.title = title
        self.intro = intro
        self.content = content

class Article(Section):
    def __init__(self, title, intro, content, number, newtag, oldtag, name, sep):
        super(Article, self).__init__(title, intro, content)
        self.number = number
        self.name = name
        self.newtag = newtag
        self.oldtag = oldtag
        # ': ' for all except for japanese where it's '\uFFA1'
        self.sep = sep
        for regs in content:
            regs.parent = self

    def depth(self):
        return 1


class TableOfContent(Section):
    def __init__(self, title, intro, content):
        super(TableOfContent, self).__init__(title, intro, content)
        self.articles = []

    def set_articles(self, articles):
        self.articles = articles

class Subsection(Section):
    def __init__(self, title, intro, content):
        super(Subsection, self).__init__(title, intro, content)

def split_rule_number(number):
    retval = []
    for elem in number:
        if elem.isdigit():
            if len(retval) > 0 and isinstance(retval[-1], int):
                retval[-1] = retval[-1] * 10 + int(elem)
            else:
                retval.append(int(elem))
        else:
            retval.append(str(elem))
            # Trick to be able to use list comparison for guidelines
            retval.append(0)
    return retval

class Rule(object):
    def __init__(self, number, text, parent):
        self.number = number.decode("utf8")
        self.list_number = split_rule_number(self.number)
        self.text = text
        self.parent = parent
        self.children = []

    def add_child(self, rule):
        self.children.append(rule)

    def depth(self):
        length = len(self.list_number)
        return length if self.list_number[-1] != 0 else length - 1

    def __lt__(self, other):
        return self.list_number < other.list_number

    def __hash__(self):
        return hash(self.number)

class Regulation(Rule):
    def __init__(self, number, text, parent):
        super(Regulation, self).__init__(number, text, parent)

class Guideline(Rule):
    def __init__(self, number, text, labelname):
        super(Guideline, self).__init__(number, text, None)
        self.labelname = labelname

    @property
    def regname(self):
        return self.number.replace('+', '')

class LabelDecl(object):
    def __init__(self, name, text):
        self.name = name
        self.text = text

    def __eq__(self, other):
        if isinstance(other, basestring):
            return self.name.__eq__(other)
        elif isinstance(other, LabelDecl):
            return self.name.__eq__(other.name)
        else:
            return False

    def __hash__(self):
        return hash(self.name)


class ASTVisitor(object):
    '''
    Basic ASTVisitor. Should never be instanciated, please use
    RegulationsASTVisitor or GuidelinesASTVisitor below as base class.
    '''
    def __init__(self):
        self.rootnode_type = None

    def visit(self, o):
        if self.rootnode_type is None:
            self.rootnode_type = o.__class__
        name = "visit" + o.__class__.__name__
        method = getattr(self, name, None)
        if method is None or not callable(method):
            print "Unable to visit " + o.__class__.__name__
            return False
        return method(o)

    def visitlist(self, o):
        retval = [self.visit(i) for i in o]
        return retval.count(False) == 0

    def visitunicode(self, s):
        return True

    def visitLabelDecl(self, label):
        return True

    def visitSection(self, section):
        return self.visit(section.intro) and self.visit(section.content)

    def visitSubsection(self, subsection):
        return self.visit(subsection.intro) and self.visit(subsection.content)

    def visitTableOfContent(self, toc):
        return True

    def visitArticle(self, article):
        return self.visit(article.intro) and self.visit(article.content)

    def visitWCADocument(self, document):
        retval = self.visit(document.sections)
        return retval and self.end_of_document(document)

    def visitWCAGuidelines(self, regs):
        return self.visitWCADocument(regs)

    def visitWCARegulations(self, regs):
        return self.visitWCADocument(regs)

    def visitRule(self, rule):
        return True

    def visitGuideline(self, reg):
        self.visitRule(reg)
        return True

    def visitRegulation(self, reg):
        return self.visitRule(reg) and self.visit(reg.children)

    def end_of_document(self, document):
        return True

# FIXME not even sure this utility belongs here
class Ruleset(ASTVisitor):
    def __init__(self):
        super(Ruleset, self).__init__()
        self.ruleset = set()

    def visitRule(self, rule):
        self.ruleset.add(rule.number)
        return True

    def get(self, ast):
        self.ruleset = set()
        if self.visit(ast):
            return self.ruleset
        else:
            return None




from wrc.sema.ast import ASTVisitor

class CGDocument(ASTVisitor):
    def __init__(self):
        super(CGDocument, self).__init__()
        self.codegen = None
        # Accepted doctypes, set by subclass
        self.doctype = None

    def emit(self, ast):
        if not isinstance(ast, self.doctype):
            print "Can't generate code for this ast(" + str(type(ast)) + ")"
            print "Accepted type : " + str(self.doctype)
            return None
        if self.visit(ast):
            return self.codegen
        else:
            return None



from wrc.sema.ast import ASTVisitor, WCAGuidelines, WCARegulations

class SemaAnalysis(ASTVisitor):
    def __init__(self):
        super(SemaAnalysis, self).__init__()
        self.errors = []
        self.warnings = []

class HierarchyCheck(SemaAnalysis):
    def __init__(self):
        super(HierarchyCheck, self).__init__()
        self.lastrule = [None]
        self.err_misplaced = "%s %s misplaced in section %s"
        self.err_precedence = "%s %s comes after %s when it shouldn't"
        self.err_reg_in_guide = "Expected a Guideline, got Regulation %s"
        self.err_guide_in_reg = "Expected a Regulation, got Guideline %s"

    def visitArticle(self, article):
        self.lastrule = [None]
        return super(HierarchyCheck, self).visitArticle(article)

    def visitRegulation(self, rule):
        retval = self.visitRule(rule)
        self.lastrule.append(None)
        retval = retval and self.visit(rule.children)
        self.lastrule.pop()
        return retval

    def visitGuideline(self, rule):
        return self.visitRule(rule)

    def visitRule(self, reg):
        last = self.lastrule[-1]
        # Here parent cannot be none, it's either another regulation or an article
        # (both have a "number" member)
        if (not reg.number.startswith(reg.parent.number) or
                (self.rootnode_type == WCARegulations and
                 reg.parent.depth() + 1 != reg.depth())):
            self.errors.append(self.err_misplaced %
                               (str(reg.__class__.__name__),
                                reg.number,
                                reg.parent.number))
        elif last and reg < last:
            self.errors.append(self.err_precedence %
                               (str(reg.__class__.__name__),
                                reg.number,
                                last.number))
        if self.rootnode_type == WCAGuidelines and reg.number.count('+') == 0:
            self.errors.append(self.err_reg_in_guide % (reg.number))

        if self.rootnode_type == WCARegulations and reg.number.count('+') != 0:
            self.errors.append(self.err_guide_in_reg % (reg.number))

        self.lastrule[-1] = reg
        return True

class LabelCheck(SemaAnalysis):
    def __init__(self):
        super(LabelCheck, self).__init__()
        self.labels = dict()
        self.err_duplicate = "Duplicate label %s"
        self.err_undefined = "Guideline %s uses label %s which is undefined"
        self.warn_unused = "Unused label %s"

    def visitLabelDecl(self, label):
        if label in self.labels:
            self.errors.append(self.err_duplicate % label.name)
        else:
            self.labels[label] = False
        return True

    def visitGuideline(self, reg):
        if reg.labelname in self.labels:
            self.labels[reg.labelname] = True
        else:
            self.errors.append(self.err_undefined % (reg.number, reg.labelname))

    def end_of_document(self, document):
        for key, value in self.labels.iteritems():
            if not value:
                self.warnings.append(self.warn_unused % key.name)
        return True





from setuptools import setup, find_packages

setup(name='wrc',
      version='1.0.0a4',
      description='Tool to build and perform checks on WCA Regulations and Guidelines',
      url='http://github.com/viroulep/wca-regulations-compiler',
      author='Philippe Virouleau',
      author_email='philippe.44@gmail.com',
      license='GPLv3',
      packages=find_packages(),
      install_requires=['ply>=3.7'],
      package_data={
          'wrc': ['data/*'],
      },
      entry_points={
          'console_scripts': ['wrc=wrc.wrc:run', 'wrc-languages=wrc.wrc:languages'],
      },
      zip_safe=False)



'''
WCALexer : tokenize a string using lex.
'''
import re
import ply.lex as lex


class WCALexer(object):

    # Token names in the Regulations and Guidelines
    tokens = (
        'TITLE',
        'VERSION',
        'TOC',
        'LABELDECL',
        'ARTICLEHEADER',
        'HEADERSEC',
        'HEADERSUBSEC',
        'TEXT',
        'PARBREAK',
        'REGULATION',
        'GUIDELINE',
        )


    t_ignore = '\t'

    def t_TITLE(self, token):
        ur'\#\s+<wca-title>(?P<title>.+)\n'
        token.value = token.lexer.lexmatch.group("title").decode("utf8")
        token.lexer.lineno += 1
        return token

    def t_VERSION(self, token):
        ur'<version>(?P<version>.+)\n'
        token.lexer.lineno += 1
        token.value = token.lexer.lexmatch.group("version").decode("utf8")
        return token

    def t_TOC(self, token):
        ur'<table-of-contents>'
        token.value = token.value.decode("utf8")
        return token

    def t_LABELDECL(self, token):
        ur'-\s<label>\s*\[(?P<label>.+?)\]\s*(?P<text>.+?)\n'
        label = token.lexer.lexmatch.group("label").decode("utf8")
        text = token.lexer.lexmatch.group("text").decode("utf8")
        token.value = (label, text)
        token.lexer.lineno += 1
        return token

    def t_ARTICLEHEADER(self, token):
        # \xef\xbc\x9a is the "fullwidth colon" used in Japanese for instance
        ur'\#\#\s+<article-(?P<number>[A-Z0-9]+)><(?P<newtag>[a-zA-Z0-9-]+)><(?P<oldtag>[a-zA-Z0-9-]+)>[ ]*(?P<name>[^\<]+?)(?P<sep>:\s|\xef\xbc\x9a)(?P<title>[^<\n]+)\n'
        number = token.lexer.lexmatch.group("number").decode("utf8")
        newtag = token.lexer.lexmatch.group("newtag").decode("utf8")
        oldtag = token.lexer.lexmatch.group("oldtag").decode("utf8")
        name = token.lexer.lexmatch.group("name").decode("utf8")
        sep = token.lexer.lexmatch.group("sep").decode("utf8")
        title = token.lexer.lexmatch.group("title").decode("utf8")
        token.value = (number, newtag, oldtag, name, title, sep)
        token.lexer.lineno += 1
        return token

    def t_HEADERSEC(self, token):
        ur'\#\#\s+(?P<title>.+?)\n'
        title = token.lexer.lexmatch.group("title").decode("utf8")
        token.value = title
        token.lexer.lineno += 1
        return token

    # This is not very flexible, but make the yacc very straightforward
    def t_HEADERSUBSEC(self, token):
        ur'\#\#\#\s+(?P<title>.+?)\n'
        title = token.lexer.lexmatch.group("title").decode("utf8")
        token.value = title
        token.lexer.lineno += 1
        return token

    def t_REGULATION(self, token):
        ur'(?P<indents>\s{4,})*-\s(?P<reg>[a-zA-Z0-9]+)\)\s*(?P<text>.+?)\n'
        indents = token.lexer.lexmatch.group("indents")
        indents = len(indents)/4 if indents else 0
        reg = token.lexer.lexmatch.group("reg").decode("utf8")
        text = token.lexer.lexmatch.group("text").decode("utf8")
        token.value = (indents, reg, text)
        token.lexer.lineno += 1
        return token

    def t_GUIDELINE(self, token):
        ur'-\s(?P<reg>[a-zA-Z0-9]+[+]+)\)\s\[(?P<label>.+?)\]\s*(?P<text>.+?)\n'
        reg = token.lexer.lexmatch.group("reg").decode("utf8")
        text = token.lexer.lexmatch.group("text").decode("utf8")
        label = token.lexer.lexmatch.group("label").decode("utf8")
        token.value = (0, reg, text, label)
        token.lexer.lineno += 1
        return token

    def t_TEXT(self, token):
        ur'(?P<text>[^<#\n ].+?)(?=\n)'
        text = token.lexer.lexmatch.group("text").decode("utf8")
        token.value = text
        return token

    def t_PARBREAK(self, token):
        ur'\n{2,}'
        token.lexer.lineno += len(token.value)
        return token


    def t_newline(self, token):
        ur'\n+'
        token.lexer.lineno += len(token.value)

    def t_error(self, token):
        print "Illegal character '%s' at line %s" % (token.value[0], token.lexer.lineno)
        token.lexer.lexerror = True
        token.lexer.skip(1)

    def lex(self):
        return lex.lex(module=self, reflags=re.UNICODE)





'''
WCAParser : turns an input string into an AST representing the
(translated) Regulations or Guidelines.
'''
import ply.yacc as yacc
from wrc.parse.lexer import WCALexer
from wrc.sema.ast import WCAGuidelines, WCARegulations, Section, Subsection,\
                      TableOfContent, Regulation, Guideline, Article, LabelDecl
from wrc.sema.check import HierarchyCheck, LabelCheck

class WCAParser(object):
    ''' Main parser class. Uses WCALexer and yacc to build the AST.'''
    def __init__(self):
        self.lexer = WCALexer().lex()
        self.tokens = WCALexer.tokens
        self.parser = yacc.yacc(module=self, debug=0, write_tables=0)
        self.doctype = WCARegulations
        self.errors = []
        self.warnings = []
        self.sema = {WCARegulations : [HierarchyCheck],
                     WCAGuidelines : [HierarchyCheck, LabelCheck]}
        self.toc = None

        # Rules hierarchy related variables
        self.prevIndent = 0
        # Contains the hierarchy at the current level, to be able to
        # know one rule's parent rule
        # Example :
        # 1 before: (None), after: (1)
        #   1.1 before: (1), after: (1, 1.1)
        #     1.1.a before: (1, 1.1), after: (1, 1.1, 1.1.a)
        #   1.2 before: (1, 1.1, 1.1.a), after: (1, 1.2)
        # 2 before: (1, 1.2), after: (2)
        # Note: it does *not* check hierarchy, ie "3.2" could be nested in "1"
        # The Sema HierarchyCheck checks this
        self.currentRule = [None]


    def parse(self, data, doctype):
        '''
        Parse an input string, and return an AST
        doctype must have WCADocument as a baseclass
        '''
        self.doctype = doctype
        self.lexer.lineno = 0
        del self.errors[:]
        del self.warnings[:]
        self.lexer.lexerror = False
        ast = self.parser.parse(data, lexer=self.lexer)
        if self.lexer.lexerror:
            ast = None
        if ast is None:
            self.errors.append("Couldn't build AST.")
        else:
            for check in self.sema[self.doctype]:
                visitor = check()
                if not visitor.visit(ast):
                    self.errors.append("Couldn't visit AST.")
                self.errors.extend(visitor.errors)
                self.warnings.extend(visitor.warnings)
        return (ast, list(self.errors), list(self.warnings))

    def _act_on_list(self, lhs):
        '''
        Act on the following rule :
            items : items item
                  | item
        '''
        lhs[0] = []
        if len(lhs) == 3:
            lhs[0] = lhs[1]
        # lhs[len(lhs)-1] may be different from lhs[-1]
        # Yacc use some internal method to get the element, see yacc.py:240
        item = lhs[len(lhs) - 1]
        if item:
            lhs[0].append(item)

    def p_content(self, content):
        '''content : TITLE opttexts VERSION opttexts sections'''
        content[0] = self.doctype(content[1], content[3], content[5])
        self.toc.set_articles([a for a in content[0].sections if isinstance(a, Article)])

    def p_texts(self, texts):
        '''texts : textlist'''
        # We want to merge back the text that we had to split in lexer
        # One \n means linebreak, \n\n changes paragraph
        texts[0] = u"\n".join(texts[1])

    def p_textlist(self, textlist):
        '''textlist : textlist text
                 | text'''
        self._act_on_list(textlist)

    def p_text(self, text):
        '''text : TEXT PARBREAK
                | TEXT
                | PARBREAK'''
        item = text[1]
        text[0] = item if item[0] != "\n" else u""
        if len(text) > 2:
            text[0] += "\n"

    def p_sections(self, sections):
        '''sections : sections section
                    | section'''
        self._act_on_list(sections)

    def p_section(self, section):
        '''section : toc
                   | article
                   | regularsec'''
        section[0] = section[1]

    def p_toc(self, toc):
        '''toc : HEADERSEC opttexts TOC opttexts'''
        toc[0] = TableOfContent(toc[1], toc[2], [])
        self.toc = toc[0]

    def p_article(self, article):
        '''article : ARTICLEHEADER opttexts rules opttexts'''
        article[0] = Article(article[1][4], article[2], article[3], article[1][0],
                             article[1][1], article[1][2], article[1][3], article[1][5])


    def p_regularsec(self, regularsec):
        '''regularsec : HEADERSEC opttexts subsections'''
        texts = []
        sections = regularsec[2]
        if len(regularsec) > 3:
            texts = regularsec[2]
            sections = regularsec[3]
        regularsec[0] = Section(regularsec[1], texts, sections)

    def p_opttexts(self, opttexts):
        '''opttexts : texts
                    | '''
        opttexts[0] = opttexts[1] if len(opttexts) > 1 else u""


    def p_subsections(self, subsections):
        '''subsections : subsections subsection
                       | subsection'''
        self._act_on_list(subsections)

    def p_subsection(self, subsection):
        '''subsection : HEADERSUBSEC texts
                      | HEADERSUBSEC texts labeldecls opttexts'''
        content = subsection[3] if len(subsection) > 3 else []
        subsection[0] = Subsection(subsection[1], subsection[2], content)

    def p_labeldecls(self, labeldecls):
        '''labeldecls : labeldecls labeldecl
                      | labeldecl'''
        self._act_on_list(labeldecls)

    def p_labeldecl(self, labeldecl):
        '''labeldecl : LABELDECL'''
        labeldecl[0] = LabelDecl(labeldecl[1][0], labeldecl[1][1])

    def p_rules(self, rules):
        '''rules : rules rule
                 | rule'''
        self._act_on_list(rules)

    def p_rule(self, rule):
        '''rule : GUIDELINE
                | REGULATION'''
        if len(rule[1]) == 4:
            # This is a guideline
            rule[0] = Guideline(rule[1][1], rule[1][2], rule[1][3])
        else:
            # This is a regulation
            indentsize = rule[1][0]
            number = rule[1][1]
            text = rule[1][2]
            parent = None

            # If we just "un"nested, shrink the current rule to our level
            if self.prevIndent > indentsize:
                self.currentRule = self.currentRule[0:indentsize+1]

            # We just added a nested level, the parent is the list's last elem
            if self.prevIndent < indentsize:
                parent = self.currentRule[-1]
            # Else, if we are nested the parent is the one before the last elem
            elif len(self.currentRule) > 1:
                parent = self.currentRule[-2]
            # Else if we are not nested, then we are a root rule and parent is none
            # (do nothing as parent is initialized to none)

            # Create the regulation node
            reg = Regulation(number, text, parent)

            # Let our parent knows he has a new child, if we don't have a parent
            # let's create an item in the article rules list
            if parent:
                parent.add_child(reg)
            else:
                rule[0] = reg

            # Unless we nested, pop and replace the last rule by ourself
            # If we added a nesting level, we just need to add ourself
            if self.prevIndent >= indentsize:
                self.currentRule.pop()
            self.currentRule.append(reg)
            self.prevIndent = indentsize

    def p_error(self, elem):
        '''Handle syntax error'''
        print "Syntax error on line " + str(self.lexer.lineno-1)
        print "Unexpected " + elem.type



import re
from wrc.sema.ast import WCAGuidelines, WCARegulations, Ruleset,\
                     Rule, LabelDecl
from wrc.codegen.cg import CGDocument

PDF_LINK = "wca-regulations-and-guidelines.pdf"
REPO_REG = "https://github.com/cubing/wca-regulations"
REPO_TRANS = "https://github.com/cubing/wca-regulations-translations"
BRANCH_REG = "official"
BRANCH_TRANS = "master"
ID_REG = "official"
ID_TRANS = "wca-regulations-translations"
TOC_ELEM = u'<li>{name}{sep}<a href="#{anchor}">{title}</a></li>\n'
H2 = u'<h2 id="{anchor}">{title}</h2>\n'
H3 = u'<h3 id="{anchor}">{title}</h3>\n'
HARTICLE = (u'<span id="{anchor}"></span><span id="{old}"></span>'
            '<h2 id="article-{anchor}-{new}"> '
            '<a href="#article-{anchor}-{new}">{name}</a>{sep}{title}'
            '</h2>\n')
PROVIDE = u"<% provide(:title, '{title}') %>\n"
TITLE = u'<h1>{title}</h1>\n'
VERSION = u'<div class="version">{version}<br/>{gitlink}</div>\n'
GITLINK = (u'[<code><a href="{repo}/tree/{branch}/{gitdir}">{identifier}</a>'
           ':<a href="{repo}/commits/{version}">{version}</a></code>]')
GUIDELINE = (u'<li id="{i}"><a href="#{i}">{i}</a>) '
             '<span class="{label} label {linked}"><a {attr}>{label}</a></span>'
             ' {text}</li>\n')
LABEL = (u'<li><span class="example {name} label label-default">{name}</span> '
         '{text}</li>\n')
REGULATION = u'<li id="{i}"><a href="#{i}">{i}</a>) {text}'
POSTREG = u'</li>\n'

# Some homemade basics non-robust md2html functions
# We could also call pando to do the conversion but it's awfully slow

def anchorizer(text):
    return text.lower().replace(" ", "-")


def special_links_replace(text):
    regOrGuide2Slots = r'([A-Za-z0-9]+)' + r'(\+*)'
    regsURL = "./"
    guidesURL = "guidelines.html"
    reference_list = [(r'regulations:article:' + regOrGuide2Slots, regsURL),
                      (r'regulations:regulation:' + regOrGuide2Slots, regsURL),
                      (r'guidelines:article:' + regOrGuide2Slots, guidesURL),
                      (r'guidelines:guideline:' + regOrGuide2Slots, guidesURL),
                     ]
    anchor_list = [(r'regulations:contents', regsURL + r'#contents'),
                   (r'guidelines:contents', guidesURL + r'#contents'),
                   (r'regulations:top', regsURL),
                   (r'guidelines:top', guidesURL),
                   (r'link:pdf', PDF_LINK),
                  ]
    retval = text
    for match, repl in reference_list:
        retval = re.sub(match, repl + r'#\1\2', retval)
    for match, repl in anchor_list:
        retval = re.sub(match, repl, retval)
    return retval

def link2html(text):
    match = r'\[([^\]]+)\]\(([^)]+)\)'
    replace = r'<a href="\2">\1</a>'
    return re.sub(match, replace, text)

def simple_md2html(text):
    retval = special_links_replace(text)
    # Create a br for every 4 spaces
    retval = re.sub(r'[ ]{4}\n', r'<br />\n', retval)
    # Do we really need this ? Help reduce the diff to only '\n' diff.
    retval = re.sub(r'"', r'&quot;', retval)
    return link2html(retval)


class WCADocumentHtml(CGDocument):
    def __init__(self, versionhash, language):
        super(WCADocumentHtml, self).__init__()
        self.codegen = u""

        is_translation = (language != "english")
        repo = REPO_TRANS if is_translation else REPO_REG
        branch = BRANCH_TRANS if is_translation else BRANCH_REG
        gid = ID_TRANS if is_translation else ID_REG
        gdir = language if is_translation else ""

        self.gitlink = GITLINK.format(repo=repo, branch=branch, identifier=gid,
                                      version=versionhash, gitdir=gdir)

    def visitWCADocument(self, document):
        # self.codegen += self.str_provide.format(title=document.title)
        self.codegen += '<div class="container">'
        self.codegen += TITLE.format(title=document.title)
        self.codegen += VERSION.format(version=document.version,
                                       gitlink=self.gitlink)
        retval = [self.visit(s) for s in document.sections]
        self.codegen += "</div>\n"
        # Codegen is a Unicode
        # FIXME do we really need ascii html entities instead of plain utf8 ?
        self.codegen = self.codegen.encode('ascii', 'xmlcharrefreplace')
        # Now codegen is a str
        # Let's provide the title in utf8, Rails should be able to handle it
        self.codegen = (PROVIDE.format(title=document.title).encode('utf8')
                        + str(self.codegen))
        return retval.count(False) == 0

    def visitlist(self, o):
        genul = len(o) > 0 and (isinstance(o[0], Rule) or isinstance(o[0], LabelDecl))
        if genul:
            self.codegen += "\n<ul>\n"
        retval = super(WCADocumentHtml, self).visitlist(o)
        if genul:
            self.codegen += "</ul>\n"
        return retval

    def visitunicode(self, u):
        if len(u) > 0:
            self.codegen += "<p>" + simple_md2html(u) + "</p>\n"
        return True

    def visitTableOfContent(self, toc):
        self.codegen += H2.format(anchor="contents",
                                  title=simple_md2html(toc.title))
        retval = super(WCADocumentHtml, self).visit(toc.intro)
        self.codegen += '<p><ul id="table_of_contents">\n'
        for article in toc.articles:
            anchorname = "article-" + article.number + "-" + article.newtag
            self.codegen += TOC_ELEM.format(name=article.name, anchor=anchorname,
                                            title=article.title, sep=article.sep)
        self.codegen += '</ul>\n</p>\n'
        return retval

    def visitSection(self, section):
        self.codegen += H2.format(anchor=anchorizer(section.title),
                                  title=section.title)
        return super(WCADocumentHtml, self).visitSection(section)

    def visitArticle(self, article):
        self.codegen += HARTICLE.format(anchor=article.number,
                                        old=article.oldtag,
                                        new=article.newtag,
                                        name=article.name,
                                        title=article.title,
                                        sep=article.sep)
        retval = super(WCADocumentHtml, self).visit(article.intro)
        retval = retval and super(WCADocumentHtml, self).visit(article.content)
        return retval

    def visitSubsection(self, subsection):
        self.codegen += H3.format(anchor=anchorizer(subsection.title),
                                  title=subsection.title)
        return super(WCADocumentHtml, self).visitSubsection(subsection)


class WCARegulationsHtml(WCADocumentHtml):
    def __init__(self, versionhash="unknown", language="english"):
        super(WCARegulationsHtml, self).__init__(versionhash, language)
        # This CG can only handle regulations
        self.doctype = WCARegulations

    def visitRegulation(self, reg):
        self.codegen += REGULATION.format(i=reg.number,
                                          text=simple_md2html(reg.text))
        retval = super(WCARegulationsHtml, self).visitRegulation(reg)
        self.codegen += POSTREG
        return retval

class WCAGuidelinesHtml(WCADocumentHtml):
    def __init__(self, regulations, versionhash="unknown", language="english"):
        super(WCAGuidelinesHtml, self).__init__(versionhash, language)
        # This CG can only handle guidelines
        self.doctype = WCAGuidelines
        self.regset = Ruleset().get(regulations)

    def visitLabelDecl(self, decl):
        self.codegen += LABEL.format(name=decl.name, text=decl.text)
        return True

    def visitGuideline(self, guide):
        reg = guide.regname
        linked = reg in self.regset
        label_class = "linked" if linked else ""
        link_attr = 'href="./#%s"' % reg
        anchor_attr = 'id="#%s"' % guide.number
        attr = link_attr if linked else anchor_attr

        self.codegen += GUIDELINE.format(i=guide.number,
                                         text=simple_md2html(guide.text),
                                         label=guide.labelname,
                                         linked=label_class,
                                         attr=attr)
        return True






import re
from wrc.sema.ast import WCADocument, WCARegulations, Ruleset,\
                     Rule, LabelDecl
from wrc.codegen.cg import CGDocument

PDF_LINK = "wca-regulations-and-guidelines.pdf"
H2 = u'\\subsection{{{title}}}\\label{{{anchor}}}\n'
H3 = u'\\subsubsection{{{title}}}\\label{{{anchor}}}\n'
TITLE = u'\\section{{{title}}}\\label{{{anchor}-top}}\n'

GUIDELINE = u'\\item \\label{{{i}}}\n {i}) [{label}] {text}\n'
LABEL = u'\\item\n {{[}}{{{name}}}{{]}} {{{text}}}\n'
REGULATION = u'\\item \\label{{{i}}} {i}) {text}\n'

CLASS = u"\\documentclass[12pt]{article}"

HEADER = """\\usepackage[top=2cm, bottom=2cm, left=2cm, right=2cm]{geometry}
\\usepackage[bookmarksopen=true]{hyperref}
\\hypersetup{pdfborderstyle={/S/U/W 1},pdfborder=0 0 1}

\\usepackage{bookmark}

\\usepackage{fancyhdr}
\\pagestyle{fancy}

\\setcounter{secnumdepth}{-1}

\\providecommand{\\tightlist}{%
  \\setlength{\\itemsep}{0pt}\\setlength{\\parskip}{0pt}}

\\title{WCA Regulations and Guidelines}
\\author{WCA Regulations Committee}
\\date{\\vspace{-1em}}

\\begin{document}

\\maketitle\n"""

BREAK = """\\newpage"""
FOOTER = """\\end{document}"""

ENCODING = {"default": [],
            "cjk": ["\\usepackage{xeCJK}", "\\setCJKmainfont{AR PL UMing CN}"],
            "hungarian": ["\\usepackage[magyar]{babel}",
                          "\\usepackage[T1]{fontenc}",
                          "\\usepackage[utf8x]{inputenc}"],
            "korean": ["\\usepackage[fallback]{xeCJK}",
                       "\\usepackage{fontspec}",
                       "\\setCJKmainfont{UnBatang}"],
            "russian": ["\\usepackage[utf8]{inputenc}", "\\usepackage[russian]{babel}"],
            "french": ["\\usepackage[french]{babel}",
                       "\\usepackage[T1]{fontenc}",
                       "\\usepackage[utf8]{inputenc}"],
            "utf8": ["\\usepackage[utf8]{inputenc}"]
           }

# Some homemade basics non-robust md2latex functions
# We could also call pandoc to do the conversion but it's awfully slow

def anchorizer(text):
    accepted = re.compile(r'[\W_ ]+')
    text = accepted.sub('', text)
    if len(text) == 0:
        text = "0"
    return text.lower().replace(" ", "-")


def special_links_replace(text):
    regOrGuide2Slots = r'([A-Za-z0-9]+)' + r'(\+*)'
    regsURL = "wca-regulations"
    guidesURL = "wca-guidelines"
    # FIXME: Instead of doing this dirty replacement we could just make the
    # regulations and guidelines labels be 'regulations:regulation:thing'!
    reference_list = [(r'regulations:article:' + regOrGuide2Slots, "reg-"),
                      (r'regulations:regulation:' + regOrGuide2Slots, ""),
                      (r'guidelines:article:' + regOrGuide2Slots, "guide-"),
                      (r'guidelines:guideline:' + regOrGuide2Slots, ""),
                     ]
    # FIXME: same here
    anchor_list = [(r'regulations:contents', regsURL + "-contents"),
                   (r'guidelines:contents', guidesURL + "-contents"),
                   (r'regulations:top', regsURL + "-top"),
                   (r'guidelines:top', guidesURL + "-top"),
                   # FIXME: very ugly (why do we even bother, they are reading it!)
                   (r'link:pdf', WCADocumentLatex.root_pdf + PDF_LINK),
                  ]
    retval = text
    for match, repl in reference_list:
        retval = re.sub(match, repl + r'\1\2', retval)
    for match, repl in anchor_list:
        retval = re.sub(match, repl, retval)
    return retval

def link2latex(text):
    match = r'\[([^\]]+)\]\((?!regulations|guidelines)([^)]+)\)'
    replace = r'\href{\2}{\1}'
    return re.sub(match, replace, text)

def anchorlink2latex(text):
    match = r'\[([^\]]+)\]\(([^)]+)\)'
    replace = r'\hyperref[\2]{\1}'
    return re.sub(match, replace, text)

def simple_md2latex(text):
    # We want to escape '#', '{', '}' as they are special chars for Latex
    text = re.sub(r'([{}#])', r'\\\1', text)
    # LaTeX will render two dashes as an em hyphen.
    # This regex replacement makes sure Megaminx notation is rendered correctly.
    text = re.sub(r'--', r'-{}-', text)
    # Convert non trailing newlines to break in paragraph
    text = re.sub(r'\n(?=.)', r'\\\\\n', text)
    text = link2latex(text)
    text = special_links_replace(text)
    return anchorlink2latex(text)


class WCADocumentLatex(CGDocument):
    root_pdf = ""
    def __init__(self, language, encoding):
        super(WCADocumentLatex, self).__init__()
        self.codegen = u""
        self.codegen = u"\n".join([CLASS,
                                   "\n".join(ENCODING[encoding]),
                                   HEADER])
        # This CG can handle both
        self.doctype = WCADocument
        self.regset = []
        self.current = None
        online_url = "https://www.worldcubeassociation.org/regulations/"
        if language != "english":
            online_url += "translations/" + language
        WCADocumentLatex.root_pdf = online_url

    def emit(self, regulations, guidelines):
        # FIXME: this override a function with a different number of arguments,
        # this is bad
        self.regset = Ruleset().get(regulations)
        retval = self.visit(regulations)
        self.codegen += BREAK
        retval = retval and self.visit(guidelines)
        self.codegen += FOOTER
        return self.codegen.encode("utf8")

    def visitWCADocument(self, document):
        self.current = document.__class__
        self.codegen += TITLE.format(anchor=anchorizer(document.title),
                                     title=document.title)
        self.codegen += document.version + "\n"
        retval = [self.visit(s) for s in document.sections]
        return retval.count(False) == 0

    def visitlist(self, o):
        genul = len(o) > 0 and (isinstance(o[0], Rule) or isinstance(o[0], LabelDecl))
        if genul:
            self.codegen += "\n\\begin{itemize}\n\\tightlist\n"
        retval = super(WCADocumentLatex, self).visitlist(o)
        if genul:
            self.codegen += "\\end{itemize}\n"
        return retval

    def visitunicode(self, u):
        if len(u) > 0:
            self.codegen += simple_md2latex(u) + "\n"
        return True

    def visitTableOfContent(self, toc):
        # Remove the contents tag from the title
        title = re.sub(r'<contents>\s*', r'', toc.title)
        prefix = "wca-regulations-" if self.current == WCARegulations else "wca-guidelines-"
        self.codegen += H2.format(anchor=prefix + "contents",
                                  title=simple_md2latex(title))
        retval = super(WCADocumentLatex, self).visit(toc.intro)
        # FIXME: do we want to output an actual table of content?
        return retval

    def visitSection(self, section):
        self.codegen += H2.format(anchor=anchorizer(section.title),
                                  title=section.title)
        return super(WCADocumentLatex, self).visitSection(section)

    def visitArticle(self, article):
        prefix = "reg-" if self.current == WCARegulations else "guide-"
        self.codegen += H2.format(anchor=prefix + article.number,
                                  title=article.sep.join([article.name,
                                                          article.title]))
        retval = super(WCADocumentLatex, self).visit(article.intro)
        retval = retval and super(WCADocumentLatex, self).visit(article.content)
        return retval

    def visitSubsection(self, subsection):
        self.codegen += H3.format(anchor=anchorizer(subsection.title),
                                  title=subsection.title)
        return super(WCADocumentLatex, self).visitSubsection(subsection)

    def visitRegulation(self, reg):
        self.codegen += REGULATION.format(i=reg.number,
                                          text=simple_md2latex(reg.text))
        retval = super(WCADocumentLatex, self).visitRegulation(reg)
        return retval

    def visitLabelDecl(self, decl):
        self.codegen += LABEL.format(name=decl.name, text=decl.text)
        return True

    def visitGuideline(self, guide):
        reg = guide.regname
        if reg in self.regset:
            label = "\\hyperref[%s]{%s}" % (reg, guide.labelname)
        else:
            label = guide.labelname

        self.codegen += GUIDELINE.format(i=guide.number,
                                         text=simple_md2latex(guide.text),
                                         label=label)
        return True


